
#include "knapsack.h"
using namespace std;

int Knapsack::Instance::size() const
{
    return Packets.size();
}

Knapsack::Selection::Selection()
{
    Value = 0;
    Weight = 0;
}

void Knapsack::Solver::Solve(const Knapsack::Instance& instance, int Depth, Knapsack::Selection& temp,
                             Knapsack::Selection& result, vector<int>& ReachableValue, vector<int>& CombinedWeights)
{
    // current choice is too heavy
    if(temp.Weight > instance.MaximumWeight)
        return;
    
    // current choice is a valid solution
    if(Depth >= instance.size())
    {
        // if current solution is better than the best previous solution
        if(temp.Value > result.Value)
            result = temp;
        return;
    }
    
    // branch and bound
    if(temp.Value+ReachableValue[Depth] <= result.Value)
        return;

    // recurse
    const Knapsack::Packet* p = &instance.Packets[Depth];
    
    temp.Choice.push_back(p);
    temp.Value += p->Value;
    temp.Weight += p->Weight;
    Solve(instance, Depth+1, temp, result, ReachableValue, CombinedWeights);
    
    temp.Choice.pop_back();
    temp.Value -= p->Value;
    temp.Weight -= p->Weight;
    
    // if temp.Weight + CombinedWeights[Depth] is <= instance.MaximumWeight, then we
    // can choose all subsequent packets without exceeding the weight-limit. Hence,
    // the first case is defenitely better than the second case, so the second case
    // need not be tested anymore.
    if(temp.Weight + CombinedWeights[Depth] > instance.MaximumWeight)
        Solve(instance, Depth+1, temp, result, ReachableValue, CombinedWeights);
}

Knapsack::Selection Knapsack::Solver::Solve(const Instance& instance)
{
    Knapsack::Selection result;
    Knapsack::Selection temp;
    
    vector<int> ReachableValue(instance.size());
    ReachableValue[instance.size()-1] = instance.Packets[instance.size()-1].Value;
    for(int i = instance.size()-2; i >= 0; --i)
        ReachableValue[i] = ReachableValue[i+1] + instance.Packets[i].Value;

    vector<int> CombinedWeights(instance.size());
    CombinedWeights[instance.size()-1] = instance.Packets[instance.size()-1].Weight;
    for(int i = instance.size()-2; i >= 0; --i)
        CombinedWeights[i] = CombinedWeights[i+1] + instance.Packets[i].Weight;

    Solve(instance, 0, temp, result, ReachableValue, CombinedWeights);
    return result;
}